@@ -29,6 +29,7 @@
 #include <fcntl.h>
 #include <errno.h>
 #include <sys/ioctl.h>
+#include <sys/mount.h>
 #include <sys/socket.h>
 #include <sys/sysmacros.h>
 #include <sys/un.h>
@@ -1062,14 +1063,16 @@ t_testbed_libc(UMockdevTestbedFixture * fixture, UNUSED_DATA)
     GError *error = NULL;
     char *path;
     char pathbuf[PATH_MAX];
-    int dirfd, fd;
+    int fd;
     struct stat st;
     uid_t uid = getuid();
 
     /* start with adding one device */
     success = umockdev_testbed_add_from_string(fixture->testbed,
 					       "P: /devices/dev1\n"
+					       "N: pd1\n"
 					       "E: SUBSYSTEM=pci\n"
+					       "E: DEVNAME=/dev/pd1\n"
 					       "A: simple_attr=1", &error);
     g_assert_no_error(error);
     g_assert(success);
@@ -1113,35 +1116,61 @@ t_testbed_libc(UMockdevTestbedFixture * fixture, UNUSED_DATA)
 
     /* openat */
 
+    int dirfd_root = open("/", O_RDONLY | O_DIRECTORY);
+    g_assert_cmpint(dirfd_root, >=, 0);
+    int dirfd_bin = open("/bin", O_RDONLY | O_DIRECTORY);
+    g_assert_cmpint(dirfd_bin, >=, 0);
+    int dirfd_dev = open("/dev", O_RDONLY | O_DIRECTORY);
+    int dirfd_sys = open("/sys", O_RDONLY | O_DIRECTORY);
+
     /* sys/ in root dir should be trapped */
-    dirfd = open("/", O_RDONLY | O_DIRECTORY);
-    g_assert_cmpint(dirfd, >=, 0);
-    fd = openat(dirfd, "sys/devices/dev1/simple_attr", O_RDONLY);
+    fd = openat(dirfd_root, "sys/devices/dev1/simple_attr", O_RDONLY);
+    if (fd < 0)
+        perror("openat");
+    g_assert_cmpint(fd, >=, 0);
+    close(fd);
+
+    /* dirfd == sys/ should be trapped */
+    fd = openat(dirfd_sys, "devices/dev1/simple_attr", O_RDONLY);
     if (fd < 0)
         perror("openat");
     g_assert_cmpint(fd, >=, 0);
     close(fd);
 
 #ifdef HAVE_OPENAT64
-    fd = openat64(dirfd, "sys/devices/dev1/simple_attr", O_RDONLY);
+    fd = openat64(dirfd_root, "sys/devices/dev1/simple_attr", O_RDONLY);
     if (fd < 0)
         perror("openat64");
     g_assert_cmpint(fd, >=, 0);
     close(fd);
 #endif
 
-    close(dirfd);
+    /* dev/ in root dir should be trapped */
+    fd = openat(dirfd_root, "dev/pd1", O_RDONLY);
+    if (fd < 0)
+        perror("openat");
+    g_assert_cmpint(fd, >=, 0);
+    close(fd);
+
+    /* dirfd == dev/ should be trapped */
+    fd = openat(dirfd_dev, "pd1", O_RDONLY);
+    if (fd < 0)
+        perror("openat");
+    g_assert_cmpint(fd, >=, 0);
+    close(fd);
 
     /* sys/ in other dir should not be trapped */
-    dirfd = open("/bin", O_RDONLY | O_DIRECTORY);
-    g_assert_cmpint(dirfd, >=, 0);
     errno = 0;
-    g_assert_cmpint(openat(dirfd, "sys", O_RDONLY), <, 0);
+    g_assert_cmpint(openat(dirfd_bin, "sys", O_RDONLY), <, 0);
+    g_assert_cmpint(errno, ==, ENOENT);
+
+    /* dev/ in other dir should not be trapped */
+    errno = 0;
+    g_assert_cmpint(openat(dirfd_bin, "pd1", O_RDONLY), <, 0);
     g_assert_cmpint(errno, ==, ENOENT);
 
 #ifdef HAVE_OPENAT64
-    g_assert_cmpint(openat64(dirfd, "sys", O_RDONLY), <, 0);
-    close(dirfd);
+    g_assert_cmpint(openat64(dirfd_bin, "sys", O_RDONLY), <, 0);
 
     errno = 0;
     g_assert_cmpint(openat(AT_FDCWD, "sys/devices", O_RDONLY), <, 0);
@@ -1149,6 +1178,39 @@ t_testbed_libc(UMockdevTestbedFixture * fixture, UNUSED_DATA)
     g_assert_cmpint(openat64(AT_FDCWD, "sys/devices", O_RDONLY), <, 0);
 #endif
 
+#if defined(__GLIBC__) && defined(HAVE_OPEN_TREE)
+    /* open_tree */
+
+    /* sys/ in root dir should be trapped */
+    fd = open_tree(dirfd_root, "sys/devices/dev1/simple_attr", O_RDONLY);
+    if (fd < 0 && errno == ENOSYS) {
+        /* can happen under valgrind or old kernels */
+        g_printf("SKIP: open_tree() not available\n");
+    } else {
+        if (fd < 0)
+            perror("openat");
+        g_assert_cmpint(fd, >=, 0);
+        close(fd);
+
+        /* dev/ in root dir should be trapped */
+        fd = open_tree(dirfd_root, "dev/pd1", O_RDONLY);
+        if (fd < 0)
+            perror("openat");
+        g_assert_cmpint(fd, >=, 0);
+        close(fd);
+
+        /* sys/ in other dir should not be trapped */
+        errno = 0;
+        g_assert_cmpint(open_tree(dirfd_bin, "sys", O_RDONLY), <, 0);
+        g_assert_cmpint(errno, ==, ENOENT);
+    }
+#endif
+
+    close(dirfd_root);
+    close(dirfd_bin);
+    close(dirfd_dev);
+    close(dirfd_sys);
+
     /* stat */
     g_assert_cmpint(stat ("/sys/bus/pci/devices", &st), ==, 0);
     g_assert_cmpint(st.st_nlink, >=, 1);
@@ -1194,7 +1256,7 @@ t_testbed_libc(UMockdevTestbedFixture * fixture, UNUSED_DATA)
     g_assert(S_ISDIR(stx.stx_mode));
 
     struct statfs buf;
-    dirfd = open("/sys", O_RDONLY | O_DIRECTORY);
+    int dirfd = open("/sys", O_RDONLY | O_DIRECTORY);
     g_assert_cmpint(dirfd, >=, 0);
     g_assert_cmpint(fstatfs(dirfd, &buf), ==, 0);
     g_assert_cmpint(buf.f_type, ==, SYSFS_MAGIC);
@@ -1280,6 +1342,22 @@ t_testbed_usb_lsusb(UMockdevTestbedFixture * fixture, UNUSED_DATA)
         g_assert(strstr(out, "Canon, Inc. PowerShot SX200 IS"));
 }
 
+#define assert_dev_zero_stat(st) \
+{ \
+    g_assert_cmpint((st)->st_uid, ==, uid); \
+    g_assert(S_ISCHR((st)->st_mode)); \
+    g_assert_cmpint((st)->st_rdev, ==, 0); \
+}
+
+#define assert_dev_zero_fstat(fd) \
+{ \
+    GStatBuf st_macro; \
+    g_assert_cmpint((fd), >, 0); \
+    g_assert_cmpint(fstat((fd), &st_macro), ==, 0); /* we don't set any dev= */ \
+    close(fd); \
+    assert_dev_zero_stat(&st_macro); \
+}
+
 static void
 t_testbed_dev_access(UMockdevTestbedFixture * fixture, UNUSED_DATA)
 {
@@ -1314,16 +1392,13 @@ t_testbed_dev_access(UMockdevTestbedFixture * fixture, UNUSED_DATA)
     /* now /dev/zero should be the mocked one */
     g_assert_cmpint(g_open("/dev/wishyouwerehere", O_RDONLY, 0), ==, -1);
     g_assert_cmpint(errno, ==, ENOENT);
+
     g_assert_cmpint(g_stat("/dev/zero", &st), ==, 0);
-    g_assert_cmpuint(st.st_uid, ==, uid);
-    g_assert(S_ISCHR(st.st_mode));
-    g_assert_cmpint(st.st_rdev, ==, 0);	/* we did not set anything */
+    assert_dev_zero_stat(&st);
 
     memset(&st, 42, sizeof st);
     g_assert_cmpint(fstatat(AT_FDCWD, "/dev/zero", &st, 0), ==, 0);
-    g_assert_cmpuint(st.st_uid, ==, uid);
-    g_assert(S_ISCHR(st.st_mode));
-    g_assert_cmpint(st.st_rdev, ==, 0);	/* we did not set anything */
+    assert_dev_zero_stat(&st);
 
 #ifdef __GLIBC__
     struct statx stx;
@@ -1400,6 +1475,24 @@ t_testbed_dev_access(UMockdevTestbedFixture * fixture, UNUSED_DATA)
     } else {
         g_printf("(Skipping O_TMPFILE test, not supported on this kernel: %m) ");
     }
+
+    /* fstat() via standard open() */
+    fd = g_open("/dev/zero", O_RDONLY, 0);
+    assert_dev_zero_fstat(fd);
+
+    /* fstat() via openat() */
+    int dirfd = g_open("/dev", O_RDONLY, 0);
+    g_assert_cmpint(dirfd, >, 0);
+    fd = openat(dirfd, "zero", O_RDONLY, 0);
+    close(dirfd);
+    assert_dev_zero_fstat(fd);
+
+    /* fstat() on duplicated fd */
+    int fd1 = g_open("/dev/zero", O_RDONLY, 0);
+    int fd2 = dup(fd1);
+    g_assert_cmpint(fd1, !=, fd2);
+    assert_dev_zero_fstat(fd1);
+    assert_dev_zero_fstat(fd2);
 }
 
 static void
@@ -1422,6 +1515,14 @@ t_testbed_add_from_string_dev_char(UMockdevTestbedFixture * fixture, UNUSED_DATA
     g_assert(S_ISCHR(st.st_mode));
     g_assert_cmpint(st.st_rdev, ==, makedev(1, 3));
 
+    /* fstat() should also work */
+    int fd = g_open("/dev/empty", O_RDONLY, 0);
+    g_assert_cmpint(fd, >, 0);
+    g_assert_cmpint(fstat(fd, &st), ==, 0);
+    close(fd);
+    g_assert(S_ISCHR(st.st_mode));
+    g_assert_cmpint(st.st_rdev, ==, makedev(1, 3));
+
     /* N: with S: should create a symlink to DEVNAME */
     g_assert(umockdev_testbed_add_from_string(fixture->testbed,
 					      "P: /devices/target\nN: target\nS: link\n"
@@ -1473,7 +1574,10 @@ t_testbed_add_from_string_dev_block(UMockdevTestbedFixture * fixture, UNUSED_DAT
     /* N: without value should create an empty dev */
     g_assert(umockdev_testbed_add_from_string(fixture->testbed,
 					      "P: /devices/empty\n"
-					      "N: empty\n" "E: SUBSYSTEM=block\n" "E: DEVNAME=/dev/empty\n", &error));
+					      "A: dev=5:4\n"
+					      "N: empty\n"
+					      "E: SUBSYSTEM=block\n"
+					      "E: DEVNAME=/dev/empty\n", &error));
     g_assert_no_error(error);
 
     g_assert(g_file_get_contents("/dev/empty", &contents, &length, &error));
@@ -1489,6 +1593,14 @@ t_testbed_add_from_string_dev_block(UMockdevTestbedFixture * fixture, UNUSED_DAT
     g_assert_cmpuint(st.st_uid, ==, getuid());
     g_assert(S_ISBLK(st.st_mode));
 
+    /* fstat() should also work */
+    int fd = g_open("/dev/empty", O_RDONLY, 0);
+    g_assert_cmpint(fd, >, 0);
+    g_assert_cmpint(fstat(fd, &st), ==, 0);
+    close(fd);
+    g_assert(S_ISBLK(st.st_mode));
+    g_assert(st.st_rdev == makedev(5, 4));
+
 #ifdef __GLIBC__
     struct statx stx;
     g_assert_cmpint(statx (AT_FDCWD, "/dev/empty", 0, STATX_TYPE|STATX_UID, &stx), ==, 0);
